<?php

/**
 * Description of OAuthAS
 *
 * @author Luis Javier Gomez Santana <luis.gomez at rediris.es>
 */
include_once('assertions/SAML2AssertionChecking.php');
include_once('assertions/SIRAssertionChecking.php');
include_once('ErrorList.php');
include_once('ClientConfiguration.php');
include_once('ServerKeys.php');
// ************ JWT ************ //
include_once('AuthzServerJWT.php');

class OAuthAS {
    const SAML2 = "urn:oasis:names:tc:SAML:2.0:assertion";
    const PAPI = "urn:mace:rediris.es:papi";

    protected $error;
    protected $debug_active;
    protected $clients;
    protected $servers;
    protected $lifetime;
    protected $assertion_checking;
    protected $scope;
    protected $assertion;
    protected $assertion_type;
    protected $client_id;
    protected $access_token;
    protected $errors;
    protected $config_dir;

    /**
     * Main constructor
     * @param String $dir  
     */
    public function __construct($dir="") {
        if (0 == strcmp($dir, "")) {
            $this->config_dir = dirname(dirname(__FILE__)) . "/config/";
        } else {
            $this->config_dir = $dir;
            $last_char = substr($dir, strlen($dir) - 1);
            if (strcmp("/", $last_char) != 0) {
                $this->config_dir .="/";
            }
        }
        $this->clients = new ClientConfiguration($this->config_dir);
        $this->error = null;
        $this->errors = new ErrorList($this->config_dir);
        $this->servers = new ServerKeys($this->config_dir);
        $this->lifetime = 3600;        
        $this->assertion_checking = null;
        $this->debug_active = false;
        $this->assertion = null;
        $this->client_id = null;
        $this->assertion_type = null;
        $this->scope = null;
        $this->access_token = null;        
    }

    /**
     *
     * @param String $message 
     */
    private function error($message) {
        if ($this->debug_active) {
            error_log("OAuth_AS: " . $message);
        }
    }

    /**
     * Function that manages the request of the app client and return an appropiate response
     * @param String $post 
     */
    public function manageRequest($post) {
        $this->error("manageRequest");
        if ($this->isValidFormatRequest($post)) {
            if ($this->isValidScope()) {
                if ($this->isValidClient()) {
                    if ($this->isValidAssertion()) {                    
                        if (strcmp($this->clients->getGenerateAccessTokenMethod($this->client_id), "AS") == 0) {
                            $this->generateAccessToken();
                        } elseif (strcmp($this->clients->getGenerateAccessTokenMethod($this->client_id), "STS") == 0) {
                            $this->generateAccessTokenSTS();
                        }
                        $this->manageASResponse();
                        $this->setLogMsg();
                    }
                }
            }
        }
        if ($this->error != null) {
            $this->manageASErrorResponse();
        }
    }

    /**
     * Function that checks if the request has a valid format
     * @param Array $search Parameters of the request
     * @return boolean True if the request has a valid format, false otherwise
     */
    private function isValidFormatRequest($search) {
        $this->error("isValidFormatRequest");
        $res = false;
        if (is_array($search)
                && array_key_exists("grant_type", $search)
                //añadido el soporte para grant_type de SAML2 urn:oasis:names:tc:SAML:2.0:assertion
                && ((strcmp($search["grant_type"], "assertion") == 0) || strcmp($search["grant_type"], "urn:oasis:names:tc:SAML:2.0:assertion"))
                && array_key_exists("assertion_type", $search)
                && ($search['assertion_type'] != "")
                && array_key_exists("assertion", $search)
                && ($search['assertion'] != "")
                && array_key_exists("client_id", $search)
                && ($search['client_id'] != "")) {
            $this->assertion_type = $search['assertion_type'];
            $this->assertion = $this->checkAssertionType($this->assertion_type, $search['assertion']);
            $this->client_id = $search['client_id'];          
            if (array_key_exists("scope", $search)) {
                $this->scope = $search['scope'];
            }
            $res = true;
        } else if (array_key_exists("grant_type", $search)
                && strcmp($search["grant_type"], "assertion") != 0) {
            $this->error = "unsupported_grant_type";
        } else {
            $this->error = "invalid_request";
        }
        return $res;
    }

    /**
     *
     * @param String $assertion_type
     * @param String $assertion
     * @return String 
     */
    private function checkAssertionType($assertion_type, $assertion) {
        if ($assertion_type == 'urn:oasis:names:tc:SAML:2.0:assertion')
            $ass = json_decode(oauthAS::b64urldecode($assertion), true);
        elseif ($assertion_type == 'urn:mace:rediris.es:papi')
            $ass = json_decode(oauthAS::b64urldecode($assertion), true);
        return $ass;
    }

    /**
     *
     * @param String $string
     * @return String 
     */
    public static function b64urldecode($string) {
        $data = str_replace(array('-', '_'), array('+', '/'), $string);
        $mod4 = strlen($data) % 4;
        if ($mod4) {
            $data .= substr('====', $mod4);
        }
        return base64_decode($data);
    }

    /**
     * Needed to avoid problems unserializing
     * @param Array $array
     * @return Array 
     */
    private function transformArray($array) {
        $this->error("transformArray");
        $aux = array();
        foreach ($array as $key => $value) {
            if (is_array($value)) {
                $key = str_replace("_dos_puntos_", ":", $key);
                $aux[$key] = $this->transformArray($value);
            } else {
                $key = str_replace("_dos_puntos_", ":", $key);
                $value = str_replace("_dos_puntos_", ":", $value);
                $aux[$key] = $value;
            }
        }
        return $aux;
    }

    /**
     *
     * @return boolean 
     */
    private function isValidScope() {
        $this->error("isValidScope");
        $res = false;
        if (array_key_exists($this->cleanScope($this->scope), $this->servers)) {
            $this->error = "invalid_scope";
        } else {
            $res = true;
        }
        return $res;
    }

    /**
     *
     * @param String $scope
     * @return boolean 
     */
    private function cleanScope($scope) {
        $pos = strpos($scope, "?");
        if ($pos === false) {
            $res = $scope;
        } else {
            $res = substr($scope, 0, $pos);
        }
        return $res;
    }

    /**
     * Funcion that checks if the client making the request is an authorized one
     * @return boolean 
     */
    private function isValidClient() {
        $this->error("isValidClient");
        $res = false;
        $auth = apache_request_headers();
        $authBasic = $auth['Authorization'];
        $filtro = '/^Basic (.*?)$/';
        $n = preg_match_all($filtro, $authBasic, $out);
        if (1 == $n) {
            $hmac = $out[1][0];
            if ($this->clients->isClient($this->client_id)) {
                if (0 == strcmp($hmac, hash_hmac("sha256", $this->client_id, $this->clients->getSecret($this->client_id)))) {
                    $res = true;                    
                    $clients = $this->clients->getClients();
                    $this->lifetime = $clients[$this->client_id]['TokenLifetime'];                    
                } else {
                    $this->error = "invalid_client";
                }
            } else {
                $this->error = "unauthorized_client";
            }
        } else {
            $this->error = "invalid_client";
        }
        return $res;
    }

    /**
     * Function that ckecks the assertion depending of the assertion type (SAML2, PAPI)
     * @return boolean 
     */
    private function isValidAssertion() {
        $this->error("isValidAssertion");
        $res = false;
        if (strcmp(OAuthAS::SAML2, $this->assertion_type) == 0) {
            $this->assertion_checking = new saml2AssertionChecking($this->cleanScope($this->scope), $this->config_dir, $this->servers->url);
        } else if (strcmp(OAuthAS::PAPI, $this->assertion_type) == 0) {
            $this->assertion_checking = new SIRAssertionChecking($this->cleanScope($this->scope), $this->config_dir);
        } else {
            $this->error = "invalid_grant";
            return $res;
        }        
        if ($this->assertion_checking->checkAssertion($this->assertion)) {
            $res = true;
        }
        if (!$res) {
            $this->error = "invalid_grant";
        }
        return $res;
    }

    /**
     * Function that generates an access token
     */
    private function generateAccessToken() {
        $this->error("generateAccessToken");
        $serverID_encoded = hash_hmac('sha256', $this->servers->getID(), $this->servers->getKey($this->scope));
        $jwt = new AuthzServerJWT("/Users/luija/Sites/html/oauth2lib/oauth_as/src/mykey.pem");        
        $jwt_encoded = $jwt->encode($this->client_id, $this->assertion_checking->getTokenInfo(), $this->scope, $this->servers->getID(), $serverID_encoded, $this->lifetime);
        $this->access_token = $jwt_encoded;
    }

    public function generateAccessTokenSTS() {
        $this->error('generateAccessTokenSTS');
        $request =
                '<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:sts="http://sts.wstrust.security.gembus.geant.net/">
         <soapenv:Header/>
        <soapenv:Body>
	<ns4:RequestSecurityToken xmlns="http://www.w3.org/2005/08/addressing" xmlns:ns2="http://schemas.xmlsoap.org/ws/2004/09/policy" xmlns:ns3="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd" xmlns:ns4="http://docs.oasis-open.org/ws-sx/ws-trust/200512/" xmlns:ns5="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd">
	<ns4:RequestType>http://docs.oasis-open.org/ws-sx/ws-trust/200512/Issue</ns4:RequestType>
	<ns4:TokenType>http://www.tokens.com/DefaultToken</ns4:TokenType>
	<ns2:AppliesTo>
		<EndpointReference>
			<Address>http://www.geant.net/sp1</Address>
		</EndpointReference>
	</ns2:AppliesTo>
	<ns4:Lifetime>
            <ns3:Created>2011-09-09T10:47:20.574Z</ns3:Created>
            <ns3:Expires>2011-09-09T12:17:20.574Z</ns3:Expires>
         </ns4:Lifetime>
            <GemToken xmlns="urn:geant:gembus:security:token:1.0:gemtoken">eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJleHAiOjEzMTU1NjcwNDA1NzQsImlhdCI6MTMxNTU2NTI0MDU3NCwiaXNzIjoidXJuOmdlYW50OmdlbWJ1czpzZWN1cml0eTpzdHM6Z2Vtc3RzIiwiYXVkIjoiaHR0cDpcL1wvd3d3LmdlYW50Lm5ldFwvc3AxIn0.BFU23F0OWXAon67HrkD4BSrF_Zdu-Dvt6p5-b_rfS-oy-2tAnmvIm9JugkfK7rPHk2AwqJhlejCWvAF8zqCXvHVwNljmAJ1HGPexSEPsZF7EpFflvZeyVh4GalBeFgqbhq_iOiZ7OFYAiiJT9mzWpXWzrZesynZJD1F5-GGWHa8qzz8qm5unosd-lXfXvDPUuG3qc0N5JBt9P4gvMVQfQ9cL55MTvCSi2hItzqUhYRmFDS7tB16LU7aynxAP1-zMSOiCdxa1vAKlWoOlC62V31aK6dNsDyKfptKhuvO1Hb91rAC5rcbl4nPnXXybRCg98lMkJC4kQeA4fk9EdEY46g</GemToken>
	</ns4:RequestSecurityToken>
         </soapenv:Body>
        </soapenv:Envelope>';
        $opt = array('Content-Type' => 'text/xml; charset=UTF-8', 'SOAPAction' => "");
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, 'http://bender.rediris.es:8197/GemSTSService/');
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_HEADER, 1);
        curl_setopt($ch, CURLINFO_HEADER_OUT, 1);
        curl_setopt($ch, CURLOPT_HTTPHEADER, $opt);
        curl_setopt($ch, CURLOPT_POSTFIELDS, $request);
        $output = curl_exec($ch);
        if (curl_errno($ch)) {
            $this->error = "Curl error requesting the Access Token";
        } else {
            $pattern = "/gemtoken\">(.*?)<\//";
            if (preg_match($pattern, $output, $match_array) == 1) {

                $this->access_token = $match_array[1];
            } else {
                $this->error = 'No se pudo obtener el token';
            }
        }
        curl_close($ch);
    }

    /**
     * Function that respond to the Client wiht the access token
     */
    private function manageASResponse() {
        $this->error("manageASResponse");
        header("HTTP/1.0 200 Ok");
        header("Content-Type: application/json");
        header("Cache-control:no-store");
        $response = array();
        $response['access_token'] = $this->access_token;
        $response['token_type'] = 'bearer';
        $response['expires_in'] = $this->lifetime;
        echo json_encode($response);
    }

    /**
     * Responds with an error if the token request is invalid or unauthorized
     */
    private function manageASErrorResponse() {
        $this->error("manageASErrorResponse");
        header("HTTP/1.0 400 Bad Request");
        header("Content-Type: application/json");
        header("Cache-control:no-store");
        $err = array();
        if (!$this->errors->hasError($this->error)) {
            $err['error'] = $this->error;
            if ($this->errors->hasDescription($this->error)) {
                $err["error_description"] = $this->errors->getDescription($this->error);
            }
            if ($this->errors->hasURI($this->error)) {
                $err["error_uri"] = $this->errors->getURI($this->error);
            }
        } else {
            $err["error"] = "Infernal Error : Error unknown";
        }
        echo json_encode($err);
    }

    //Simple Getter
    public function getError() {
        return $this->error;
    }

    public function getConfig_dir() {
        return $this->config_dir;
    }

    public function setConfig_dir($config_dir) {
        $this->config_dir = $config_dir;
    }

    private function setLogMsg() {
        $this->error("setLogMsg");
        $string = "Token Request";
        $array = array("client_id" => $this->client_id, "scope" => $this->scope, "date" => date(DATE_RFC822), "assertion" => serialize($this->assertion));
        $string.= ": " . json_encode($array) . "\n";
        file_put_contents('/tmp/oauth_access.txt', $string . PHP_EOL, FILE_APPEND | LOCK_EX);
    }

}

?>
